program Gr3D;

{
  Conjunto de Rotinas basicas para operacoes com pontos em 3D.
}

type 

     Vector4 = array[1..4] of real;

     PointPtr = ^Point;

     Point = record
               next: PointPtr;
               case boolean of
               true:(x,y,z,w: real);
               false:(v: Vector4)
             end;

     TrMat= array[1..4,1..4] of real;

     VProcType = procedure (P:Point);

procedure MakeId(var M: TrMat);
{
  Constroi uma matriz identidade
}
var i,j: integer;
begin
  for i:=1 to 4 do 
    begin
      for j:=1 to 4 do M[i,j]:=0;
      M[i,i]:=1
    end
end;  

procedure PMedio(P,Q: Point; var M: Point);
var i: integer;
begin
  for i:=1 to 4 do
    with M do v[i]:=(P.v[i]+Q.v[i])/2;
end;

procedure Insert(var L: PointPtr; P: PointPtr);
{
  Insere um ponto P numa lista de pontos L
}
begin
  P^.next:=nil;
  if L = nil then L:=P else Insert(L^.next,P)
end;

procedure Append(var L1,L2: PointPtr);
{
  faz L1 <- L1 "+" L2;
      L2 <-nil;
}
var P: PointPtr;
begin
  while L2 <> nil do
    begin
      P:=L2^.next; Insert(L1,L2); L2:=P
    end
end;


procedure InsertP(var L: PointPtr; xx,yy,zz: Real);
{
  Cria um ponto P a partir das coordenadas <x,y,z> e
  o insere numa lista.
}
var P: PointPtr;
begin
  New(P);
  with P^ do
    begin
      x:=xx; y:=yy; z:=zz; w:=1; next:=nil
    end;
  Insert(L,P)
end;

procedure Close(L: PointPtr);
{
  "fecha" uma curva representada por uma lista de pontos
}
begin
  if L <> nil then with L^ do InsertP(L,x,y,z)
end;

function Count(L: PointPtr): integer;
{
  conta o numero de pontos numa lista
}
var i: integer;
begin
  i:=0;
  while L<>nil do begin inc(i); L:=L^.next end;
  Count:=i
end;

procedure Disp(Var L: PointPtr);
{
  Libera a memoria ocupada por uma lista de pontos
}
begin
  if L <> nil then 
    begin
      Disp(L^.next);
      Dispose(L)
    end
end;

procedure WrList(P: PointPtr; s: string);
{
  Escreve na saida padrao uma lista de pontos no
  formato do POVRay
}
begin
  writeln('// ',s);
  while P<> nil do with P^ do
    begin
      writeln('<',x:9:4,',',y:9:4,',',z:9:4,'>');
      P:=P^.next
    end;
  writeln('// ***************')
end;

function SProd(P: PointPtr; var M: TrMat): PointPtr;
{
  Aplica a transformacao M numa lista P, de pontos.
  Retorna a lista de pontos transformada (a lista original
  nao e' alterada).
}
var i,j: integer; S: real; L,Q: PointPtr;
begin
  L:=nil;
  while P<>nil do
    begin
      New(Q);
      for i:=1 to 4 do
        begin 
          S:=0;
          for j:=1 to 4 do with P^ do S:=S+v[j]*M[i,j];
          Q^.v[i]:=S
        end;
      Insert(L,Q); P:=P^.next;
    end;
  SProd:=L
end;


procedure MProd(var A,B,C: TrMat);
{
  Produto de duas matrizes 4x4
}
var i,j,k: integer; S: real;
begin
  for i:=1 to 4 do
    for j:=1 to 4 do
      begin
        S:=0;
        for k:=1 to 4 do S:=S+A[i,k]*B[k,j];
        C[i,j]:=S
      end
end;

function XRot(P: PointPtr; Angle: integer): PointPtr;
{
  Aplica uma rotacao em torno do eixo X a uma lista de
  pontos.
}
var M: TrMat;
begin
  MakeId(M);
  M[2,2]:=cos((-Angle/360)*2*PI); M[2,3]:=sin((-Angle/360)*2*PI);
  M[3,2]:=-M[2,3];  M[3,3]:=M[2,2];
  XRot:=SProd(P,M)
end;

function YRot(P: PointPtr; Angle: integer): PointPtr;
{
  Aplica uma rotacao em torno do eixo Y a uma lista de
  pontos.
}
var M: TrMat;
begin
  MakeId(M);
  M[1,1]:=cos((-Angle/360)*2*PI); M[1,3]:=sin((-Angle/360)*2*PI);
  M[3,1]:=-M[1,3];  M[3,3]:=M[1,1];
  YRot:=SProd(P,M)
end;

function ZRot(P: PointPtr; Angle: integer): PointPtr;
{
  Aplica uma rotacao em torno do eixo Z a uma lista de
  pontos.
}
var M: TrMat;
begin
  MakeId(M);
  M[1,1]:=cos((-Angle/360)*2*PI); M[1,2]:=sin((-Angle/360)*2*PI);
  M[2,1]:=-M[1,2];  M[2,2]:=M[1,1];
  {write('cos:',M[1,1]:9:4,'  sin:',M[1,2]:9:4); readln;}
  ZRot:=SProd(P,M)
end;

function Scale(P: PointPtr; xs,ys,zs: real): PointPtr;
{
  Aplica uma transformacao de escala a uma lista de pontos
 }
var M: TrMat;
begin
  MakeId(M);
  M[1,1]:=xs; M[2,2]:=ys; M[3,3]:=zs;
  Scale:=SProd(P,M)
end;

function Trans(P:PointPtr; dx,dy,dz: real): PointPtr;
{
  Aplica uma transformacao de translacao a uma lista de pontos
}
var M:TrMat;
begin
  MakeId(M);
  M[1,4]:=dx; M[2,4]:=dy; M[3,4]:=dz;
  Trans:=SProd(P,M)
end;

function Bezier(L: PointPtr; n: integer): PointPtr;
{
  Calcula a curva de Bezier a partir de 4 pontos de controle,
  na lista L, de Pontos. Devolve a lista com n pontos
  correspondentes `a curva.
}
var MyP,MyL: PointPtr;
    i,j: integer; t,t1,delta: real;
    cp1,cp2,cp3,cp4: Vector4; {mantem as coordenadas dos pontos de controle }

begin
  MyL:=nil;
  if Count(L) <> 4 then begin
                          writeln('Erro em Bezier (Count <> 4)');
                          halt
                        end;
  cp1:=L^.v; L:=L^.next;
  cp2:=L^.v; L:=L^.next;
  cp3:=L^.v; L:=L^.next;
  cp4:=L^.v;
  New(MyP);
  MyP^.v := cp1; { primeiro ponto da curva }
  Insert(MyL,MyP);
  delta:= 1.0 / n;
  for i:=2 to n do
    begin
      t:=i*delta; t1:=1.0-t;
      New(MyP);
      with MyP^ do for j:=1 to 4 do
                     v[j] :=  cp1[j]*t1*t1*t1
                            + cp2[j]*3.0*t*t1*t1
                            + cp3[j]*3.0*t*t*t1
                            + cp4[j]*t*t*t;
      Insert(MyL,MyP);
    end;
  Bezier:=MyL
end;

function Dist(P1,P2: PointPtr): real;
{
  calcula a distncia entre dois pontos
}
begin
  with P1^ do
    Dist:= sqrt( sqr(x-P2^.x)+sqr(y-P2^.y)+sqr(z-P2^.z))
end;

function DistTotal(L: PointPtr): real;
{
  Calcula a soma das distncias entre os pontos
  de uma curva fechada.
}
var D: real;
begin
  D:=0;
  if L <> nil then
    while L^.next <> nil do
      begin
        D:=D+Dist(L,L^.next); L:=L^.next
      end;
  DistTotal:=D
end;


procedure AvancaAux(P1,P2: PointPtr; d: real; var P: point);
{
  determina as coordenadas do ponto P, distante d de P1,
  na direo de P2 (d <=Dist(P1,P2) )
}
var f,dd: real; Q,R: Point; Pa,Pb: PointPtr;
begin
  dd:=Dist(P1,P2);
  if dd < 10E-20 then begin writeln('AvancaAux- Erro:  dd ~= 0') end;
  f:=d/Dist(P1,P2);
  if f > 1 then begin writeln('AvancaAux - Erro: d > Dist(P1,P2)'); readln end;
  Q.v:=P1^.V; Pa:=@Q; Pb:=@R; R.v:=P2^.v; R.next:=nil;
  with Q do
    begin
      next:=nil;
      Pb:=Trans(Pb,-x,-y,-z);
      Pa:=Scale(Pb,f,f,f);
      Disp(Pb);
      Pb:=Trans(Pa,x,y,z);    Disp(Pa);
    end;
  P.v:=Pb^.v; Disp(Pb);
end;

procedure Avanca(var P: PointPtr; var Q: Point; d: real);
{
  Determina as coordenadas do ponto Q cuja "distancia" do
  ponto P na curva  d, e "avana" P para o ponto seguinte
  a Q na curva. Se o valor final de P  nil, as coordenadas
  de Q so invlidas.
}
var N: PointPtr;
begin
  if d > DistTotal(P) then writeln('Avanca - d > DistTotal - erro');
  if P <> nil then
    begin
      N:=P^.next;
       while (N<>nil) and (d > Dist(P,N)) do
         begin
           d:=d-Dist(P,N);
           P:=N; N:=P^.next
         end;
         if P <> nil then  AvancaAux(P,N,d,Q);
      P:=N;
    end;
end;

procedure NPassos(L: PointPtr; n: integer; Visita: VProcType);
{
  "Anda" n passos equidistantes na curva fechada
  representada pela lista L.
}
var D: real; P: Point; var i: integer;
begin
  D:=DistTotal(L)/n;
  P:=L^; Visita(P);
  for i:=1 to n do
    begin
      Avanca(L,P,d);
      Visita(P);
    end
end;

{$F+ }
procedure WrSph(P: Point);
begin
  with P do writeln('sphere{ <',x:7:3,',',y:7:3,',',z:7:3,'> .8  1 pigment {Pink} }');
end;

procedure WrSpheres(L:PointPtr);
begin
  NPassos(L,30,WrSph);
end;



procedure WrCones(L:PointPtr; r1,r2: real);
var P: PointPtr;
begin
  writeln('#declare no3 = union {');
  writeln('                  #declare r1 =',r1:7:3);
  writeln('                  #declare r2 =',r2:7:3);
  writeln('// ****** Lista dos cones ');
  while L <> nil do with L^ do
    begin
      P:=L^.next;
      if P <> nil then
        begin
          writeln(' cone { <',x: 7:3,',',y:7:3,',',z:7:3,'>, r1 ');
          with P^ do
          writeln('        <',x: 7:3,',',y:7:3,',',z:7:3,'>, r2 ');
          writeln('      }');
        end;
      L:=P;
    end;
  writeln('// ***********************');
  writeln('}');
end;

procedure WrBolas(L:PointPtr; r: real);
var L1,L2,Laux: PointPtr;
begin
  Laux:=XRot(L,90); L1:=ZRot(Laux,90);  Disp(Laux);
  Laux:=XRot(L1,90); L2:=ZRot(Laux,90); Disp(Laux);
  writeln('#declare no3 = blob {');
  writeln('                  #declare r =',r:7:3);
  writeln('                  threshold 1.0');
  writeln('// ****** Lista das esferas ');
  WrSpheres(L);
  WrSpheres(L1);
  WrSpheres(L2);
  writeln('      finish { phong 1 }');
  writeln('// ***********************');
  writeln('}');
end;

(*********************************** cilindros ****************)
var OldP: Point;

procedure WrCylinder(P: Point);
var Q: Point;
begin
  if OldP.next <> nil then
    begin
      { AvancaAux(@OldP,@P,0.5*Dist(@OldP,@P),Q); ponto mdio }
      PMedio(OldP,P,Q);
      with OldP do write('cylinder{ <',x:7:3,',',y:7:3,',',z:7:3,'>,');
      with Q  do writeln(' <',x:7:3,',',y:7:3,',',z:7:3,'>, 0.7 open }');
    end;
  OldP:=P;
  OldP.next:=@P;
end;

procedure WrCylinders(L:PointPtr);
begin
  OldP.next:=nil;
  NPassos(L,30,WrCylinder);
end;

procedure WrCyls(L:PointPtr);
var L1,L2,Laux: PointPtr;
begin
  Laux:=XRot(L,90); L1:=ZRot(Laux,90);  Disp(Laux);
  Laux:=XRot(L1,90); L2:=ZRot(Laux,90); Disp(Laux);
  writeln('#declare no3 = union {');
  writeln('// ****** Lista dos cilindros ');
  WrCylinders(L);
  WrCylinders(L1);
  WrCylinders(L2);
  (* writeln('      finish { phong 1 }'); *)
  writeln('// ***********************');
  writeln('}');
end;

(***********************************************************)

procedure FazNo;
const a:real=1.0; { largura do cubo }
      k: real=8; { distancia para os pontos de controle }
      n: integer = 128;
var L1,L2: PointPtr;
begin
  L1:=nil;
  { insere em L1 os 4 pontos de controle, criados a partir de um cubo }
  InsertP(L1,-a,a,0);
  InsertP(L1,-a,a,k);
  InsertP(L1,a,-k,-a);
  InsertP(L1,a, 0,-a);
  L2:=Bezier(L1, n);
  { WrCones(L2,0.6,0.65); }
  { WrBolas(L2,0.9) }
    WrCyls(L2)
end;

(**************  No de 5 lacadas ***************)

procedure WrP(L: PointPtr);
{
  Escreve uma lista de pontos no formato do POVRay
}
begin
  while L <> nil do with L^ do
    begin
      writeln(', <',x:7:3,',',y:7:3,',',z:7:3,'>');
      L:=next
    end
end;

procedure FazNo5;
const k=1.5; wx=1.8; wy=1.4; d=1.3;
var L1, L2, L3, L4, Laux: PointPtr;
begin
  L1:=nil;
  InsertP(L1,d,0,0);
  InsertP(L1,wx,-wy,0);
  InsertP(L1,wx, wy,0);
  Close(L1);
  Laux:= Trans(L1,0,0,k);
  L2:=ZRot(Laux,0); Disp(Laux);
  Laux:= Trans(L1,0,0,{-}2*k);
  L4:= ZRot(L2,72*3);
  L3:= ZRot(Laux,72*3);
  WrP(L2);
  WrP(L1);
  WrP(L3);
  WrP(L4);
end;

begin { main }
  FazNo5;
end.

(*  teste *)
const
   S=0.866025;
   S3_3=1.0;
   S1_3=1.0/3.0;
   S2_3=2.0/3.0;
   H3_3=1.0;
   H1_3=1.0/3.0;
   H2_3=2.0/3.0;
   X1= S*(3.0/3.0);
   X2= S*(4.0/3.0);
   X3= S*(5.0/3.0);
   X4= S*(6.0/3.0);
   Y1=0.5;
   Y2=Y1+1.0/6.0;
   Y3=Y1+2.0/6.0;
   Y4=Y1+3.0/6.0;
   YM=2.0;

   RR=1.0;
   GG=1.2;
   BB=1.0;

var L1,L2,L3: PointPtr;
begin
(* teste 1
   L1:=nil;
   InsertP(L1,1,0,0);
   InsertP(L1,0,1,0);
   InsertP(L1,1,1,0);
   L2:=ZRot(L1,45);
   WrLIst(L1,'Lista Original:');
   WrList(L2,'Rodou 45');
   L3:=ZRot(L2,-45);
   WrList(L3,'Volta 45');
   readln
*)
(* test2
   InsertP(L1,-X4,Y4, 0);
   InsertP(L1,-S3_3,YM,-H3_3);
   InsertP(L1,0.00,YM, 0);
   InsertP(L1,X1,Y1, 0);
   InsertP(L1,-X3,Y3, 0);
   InsertP(L1,-S2_3,YM,-H2_3);
   InsertP(L1,S1_3,YM, H1_3);
   InsertP(L1,X2,Y2, 0);
   InsertP(L1,-X2,Y2, 0);
   InsertP(L1,-S1_3,YM,-H1_3);
   InsertP(L1,S2_3,YM, H2_3);
   InsertP(L1,X3,Y3, 0);
   InsertP(L1,-X1,Y1, 0);
   InsertP(L1,0.00,YM, 0.00);
   InsertP(L1,S3_3,YM, H3_3);
   InsertP(L1,X4,Y4, 0);
   WrList(L1,'1/3 da fatia original'); readln;
   L2:=ZRot(L1,120);
   WrList(L2,' Segunda fatia'); readln;
   L3:=ZRot(L1, 240);
   WrList(L3,' Terceira fatia'); readln;
*)
(* teste 3 - Bezier *)

   L1:=nil;
   InsertP(L1,1,0,1);
   InsertP(L1,1,1,1);
   InsertP(L1,1,2,1);
   InsertP(L1,1,3,1);
   L2:=Bezier(L1,8);
   WrList(L2,' Bezier p/ 8 pontos');
end.