unit ListOp;
{
   Operacoes basicas com listas de pontos;
}

interface uses PointOp;

type VProcType = procedure (P: tPoint);

procedure lInsertPoint(var L: tPointPtr; P: tPointPtr);
procedure lInsertNewPoint(var L: tPointPtr; xx,yy,zz: real);
function  lCopy(var L:tPointPtr): tPointPtr;
procedure lTrApply(L:tPointPtr; var M: TrMat);
procedure lXRotate(L:tPointPtr; a: integer);
procedure lYRotate(L:tPointPtr; a: integer);
procedure lZRotate(L:tPointPtr; a: integer);
procedure lTranslate(L:tPointPtr; dx,dy,dz: real);
procedure lScale(L: tPointPtr; sx,sy,sz: real);
procedure lWrList(var f: text; L: tPointPtr);
function  lCount(L: tPointPtr): integer;
procedure lAppend(var L1,L2: tPointPtr);
procedure lClose(L: tPointPtr);
procedure lDisp(Var L: tPointPtr);
function  Bezier(L: tPointPtr; n: integer): tPointPtr;
function  DistTotal(L: tPointPtr): real;
procedure Avanca(var P: tPointPtr; var Q: tPoint; d: real; var f: text);
procedure NPassos(L: tPointPtr; n: integer; Visita: VProcType; var f:text);


implementation

procedure lInsertPoint(var L: tPointPtr; P: tPointPtr);
{
  Insere um ponto ao final de uma lista
}
begin
  P^.next:=nil;
  if L=nil then L:=P else lInsertPoint(L^.next,P)
end;

procedure lInsertNewPoint(var L: tPointPtr; xx,yy,zz: real);
{
  Insere um ponto ao final de uma lista (a partir das coordenadas )
}
var P: tPointPtr;
begin
  new(P,NewPoint(xx,yy,zz));
  lInsertPoint(L,P)
end;

function  lCopy(var L:tPointPtr): tPointPtr;
{
  Copia uma lista de pontos.
}
var P: tPointPtr;
begin
  if L=nil then lCopy:=L
    else with L^do
           begin
             new(P,NewPoint(C.x,C.y,c.Z));
             P^.next:=lCopy(next);
             lCopy:=P
           end
end;

procedure lTrApply(L:tPointPtr; var M: TrMat);
{
  Aplica uma transformacao M aos pontos de uma lista L
}
begin
  if L <> nil then with L^ do
    begin
      TrApply(M); lTrApply(next,M)
    end
end;

procedure lXRotate(L:tPointPtr; a: integer);
{
  "roda" os pontos de uma lista L de um angulo a em torno
  do eixo x
}
begin
  if L<>nil then with L^ do
    begin
      XRotate(a); lXRotate(Next,a)
    end
end;

procedure lYRotate(L:tPointPtr; a: integer);
{
  "roda" os pontos de uma lista L de um angulo a em torno
  do eixo y
}
begin
  if L<>nil then with L^ do
    begin
      YRotate(a); lYRotate(Next,a)
    end
end;

procedure lZRotate(L:tPointPtr; a: integer);
{
  "roda" os pontos de uma lista L de um angulo a em torno
  do eixo z
}
begin
  if L<>nil then with L^ do
    begin
      ZRotate(a); lZRotate(Next,a)
    end
end;

procedure lTranslate(L:tPointPtr; dx,dy,dz: real);
{
  Aplica uma operacao de "translacao"  aos pontos
  de uma lista L
}
begin
  if L<>nil then with L^ do
    begin
      Translate(dx,dy,dz); lTranslate(next,dx,dy,dz)
    end
end;

procedure lScale(L: tPointPtr; sx,sy,sz: real);
{
  Aplica uma operacao de escala aos pontos de uma
  lista L
}
begin
  if L<>nil then with L^ do
    begin
      Scale(sx,sy,sz); lScale(next,sx,sy,sz)
    end
end;

procedure lWrList(var f: text; L: tPointPtr);
{
  Escreve os pontos de uma lista.
}
begin
  if L<>nil then with L^ do
    begin
      WrPoint(f); lWrList(f,next)
    end
end;

function  lCount(L: tPointPtr): integer;
{
  Devolve o numero de pontos de uma lista.
}
begin
  if L=nil then lCount:=0 else lCount:=1 + lCount(L^.next)
end;

procedure lAppend(var L1,L2: tPointPtr);
{
 Concatena duas listas L1 e L2:
  faz L1 <- L1 "+" L2;
      L2 <-nil;
}
var P: tPointPtr;
begin
  while L2 <> nil do
    begin
      P:=L2^.next; lInsertPoint(L1,L2); L2:=P
    end
end;

procedure lClose(L: tPointPtr);
{
  "fecha" uma curva representada por uma lista de pontos
}
var P: tPointPtr;
begin

  if L <> nil then
    begin
      with L^.C do new(P,NewPoint(x,y,z));
      lInsertPoint(L,P)
    end;
end;

procedure lDisp(Var L: tPointPtr);
{
  Libera a memoria ocupada por uma lista de pontos
}
begin
  if L <> nil then 
    begin
      lDisp(L^.next);
      Dispose(L)
    end
end;

function Bezier(L: tPointPtr; n: integer): tPointPtr;
{
  Calcula a curva de Bezier a partir de 4 pontos de controle,
  na lista L, de Pontos. Devolve a lista com n pontos
  correspondentes `a curva.
}
var MyP,MyL: tPointPtr;
    i,j: integer; t,t1,delta: real;
    cp1,cp2,cp3,cp4: tCoord; {mantem as coordenadas dos pontos de controle }

begin
  MyL:=nil;
  if lCount(L) <> 4 then begin
                          writeln('Erro em Bezier (Count <> 4)');
                          halt
                        end;
  cp1:=L^.C; L:=L^.next;
  cp2:=L^.C; L:=L^.next;
  cp3:=L^.C; L:=L^.next;
  cp4:=L^.C;
  with cp1 do New(MyP,NewPoint(x,y,z)); { primeiro ponto da curva }
  lInsertPoint(MyL,MyP);
  delta:= 1.0 / n;
  for i:=2 to n do
    begin
      t:=i*delta; t1:=1.0-t;
      new(MyP,NewPoint(0,0,0));
      with MyP^.C do for j:=1 to 4 do
                     v[j] :=  cp1.v[j]*t1*t1*t1
                            + cp2.v[j]*3.0*t*t1*t1
                            + cp3.v[j]*3.0*t*t*t1
                            + cp4.v[j]*t*t*t;
      lInsertPoint(MyL,MyP);
    end;
  Bezier:=MyL
end;

function DistTotal(L: tPointPtr): real;
{
  Calcula a soma das distncias entre os pontos
  de uma curva.
}
var D: real;
begin
  D:=0;
  if L <> nil then
    while L^.next <> nil do
      begin
        D:=D+Dist(L^,L^.next^); L:=L^.next
      end;
  DistTotal:=D
end;

procedure AvancaAux(P1,P2: tPointPtr; d: real; var P: tPoint; var ff: text);
{
  determina as coordenadas do ponto P, distante d de P1,
  na direcao de P2 (d <=Dist(P1,P2) )
}
var f,dd: real; R: tPoint;
begin
  dd:=Dist(P1^,P2^);
  if dd < 10E-20 then begin writeln(ff,'AvancaAux- Erro:  dd ~= 0') end;
  f:=d/Dist(P1^,P2^);
  if f > 1 then begin writeln(ff,'AvancaAux - Erro: d > Dist(P1,P2)'); readln end;
  with P2^.C do R.NewPoint(x,y,z);
  with P1^.C do
    begin
      R.Translate(-x,-y,-z);
      R.Scale(f,f,f);
      R.Translate(x,y,z);
    end;
  P.C:=R.C; P.next:=nil;
end;

procedure Avanca(var P: tPointPtr; var Q: tPoint; d: real; var f: text);
{
  Determina as coordenadas do ponto Q cuja "distancia" do
  ponto P na curva  d, e "avana" P para o ponto seguinte
  a Q na curva. Se o valor final de P  nil, as coordenadas
  de Q sao invlidas.
}
var N: tPointPtr;
begin
  if d > DistTotal(P) then writeln(f,'Avanca - d > DistTotal - erro');
  if P <> nil then
    begin
      N:=P^.next;
       while (N<>nil) and (d > Dist(P^,N^)) do
         begin
           d:=d-Dist(P^,N^);
           P:=N; N:=P^.next
         end;
         if P <> nil then  AvancaAux(P,N,d,Q,f);
      P:=N;
    end;
end;

procedure NPassos(L: tPointPtr; n: integer; Visita: VProcType; var f:text);
{
  "Anda" n passos equidistantes na curva fechada
  representada pela lista L. A cada "passo"  chama o
  procedimento Visita.
}
var D: real; P: tPoint; var i: integer;
begin
  D:=DistTotal(L)/n;
  with L^.C do P.NewPoint(x,y,z); Visita(P);
  for i:=1 to n do
    begin
      Avanca(L,P,d,f);
      Visita(P);
    end
end;

end.






