unit PointOp;

{
  Conjunto de Rotinas basicas para operacoes com pontos em 3D.
}

interface

type 

     tVector4 = array[1..4] of real;
     tCoord = record
                case boolean of
                  true:(x,y,z,w: real);
                  false:(v: tVector4)
              end;


     TrMat= array[1..4,1..4] of real;

     tPointPtr = ^ tPoint;
     tPoint = object
                next: tPointPtr;
                C   : tCoord;
                constructor NewPoint(xx,yy,zz: real);
                procedure Copy(var P: tPoint);
                procedure TrApply(var M: TrMat);
                procedure XRotate(a: integer);
                procedure YRotate(a: integer);
                procedure ZRotate(a: integer);
                procedure Translate(dx,dy,dz: real);
                procedure Scale(sx,sy,sz: real);
                procedure WrPoint(var f: text);
              end;


procedure PMedio(P,Q: tPoint; var M: tPoint);
function Dist(P1,P2: tPoint): real;

procedure MakeId(var M: TrMat);
procedure MProd(var A,B,C: TrMat);


implementation


procedure MakeId(var M: TrMat);

{
  Constroi uma matriz identidade
}
var i,j: integer;
begin
  for i:=1 to 4 do 
    begin
      for j:=1 to 4 do M[i,j]:=0;
      M[i,i]:=1
    end
end;  

procedure MProd(var A,B,C: TrMat);
{
  Produto de duas matrizes 4x4
}
var i,j,k: integer; S: real;
begin
  for i:=1 to 4 do
    for j:=1 to 4 do
      begin
        S:=0;
        for k:=1 to 4 do S:=S+A[i,k]*B[k,j];
        C[i,j]:=S
      end
end;

procedure PMedio(P,Q: tPoint; var M: tPoint);
var i: integer;
begin
  M.NewPoint(0,0,0);
  for i:=1 to 4 do
    with M.C do v[i]:=(P.C.v[i]+Q.C.v[i])/2;
end;

function Dist(P1,P2: tPoint): real;
{
  calcula a distncia entre dois pontos
}
begin
  with P1.C do
    Dist:= sqrt( sqr(x-P2.C.x)+sqr(y-P2.C.y)+sqr(z-P2.C.z))
end;

constructor tPoint.NewPoint(xx,yy,zz: real);
begin
  with C do
    begin
      x:=xx; y:=yy; z:=zz; w:=1; next:=nil
    end;
end;

procedure tPoint.Copy(var P: tPoint);
{
  copia as coordenadas de um ponto.
}
begin
  P.NewPoint(c.x,c.y,c.z)
end;


procedure tPoint.TrApply(var M: TrMat);
var i,j: integer; S: real; v4:tVector4;
begin
  for i:=1 to 4 do
    begin
      S:=0;
      for j:=1 to 4 do with C do S:=S+v[j]*M[i,j];
      v4[i]:=S
    end;
  C.v:=v4
end;



procedure tPoint.XRotate(a: integer);
var M: TrMat;
begin
  MakeId(M);
  M[2,2]:=cos((-a/360)*2*PI); M[2,3]:=sin((-a/360)*2*PI);
  M[3,2]:=-M[2,3];  M[3,3]:=M[2,2];
  TrApply(M)
end;

procedure tPoint.YRotate(a: integer);
var M: TrMat;
begin
  MakeId(M);
  M[1,1]:=cos((-a/360)*2*PI); M[1,3]:=sin((-a/360)*2*PI);
  M[3,1]:=-M[1,3];  M[3,3]:=M[1,1];
  TrApply(M)
end;

procedure tPoint.ZRotate(a: integer);
var M: TrMat;
begin
  MakeId(M);
  M[1,1]:=cos((-a/360)*2*PI); M[1,2]:=sin((-a/360)*2*PI);
  M[2,1]:=-M[1,2];  M[2,2]:=M[1,1];
  TrApply(M)
end;

procedure tPoint.Translate(dx,dy,dz: real);
var M:TrMat;
begin
  MakeId(M);
  M[1,4]:=dx; M[2,4]:=dy; M[3,4]:=dz;
  TrApply(M)
end;

procedure tPoint.Scale(sx,sy,sz: real);
var M: TrMat;
begin
  MakeId(M);
  M[1,1]:=sx; M[2,2]:=sy; M[3,3]:=sz;
  TrApply(M)
end;


procedure tPoint.WrPoint(var f: text);
begin
  write(f,' <',C.x:7:3,',',C.y:7:3,',',C.z:7:3,'>');
end;

end.